use std::{collections::HashSet, fs::read_to_string, io::Write, ops::RangeInclusive};

struct Sensor {
    x: i64,
    y: i64,
    x_beacon: i64,
    y_beacon: i64,
    distance: u64,
}

impl From<&str> for Sensor {
    fn from(value: &str) -> Self {
        let [x_sensor, y_sensor, x_beacon, y_beacon] = value
            .chars()
            .filter(|c| c.is_ascii_digit() || c.is_whitespace() || *c == '-')
            .collect::<String>()
            .split_whitespace()
            .map(|n| n.parse::<i64>().unwrap())
            .collect::<Vec<_>>()[..] else {panic!()};
        Self {
            x: x_sensor,
            y: y_sensor,
            x_beacon,
            y_beacon,
            distance: x_sensor.abs_diff(x_beacon) + y_sensor.abs_diff(y_beacon),
        }
    }
}

fn main() {
    let input = (
        "Sensor at x=2, y=18: closest beacon is at x=-2, y=15
Sensor at x=9, y=16: closest beacon is at x=10, y=16
Sensor at x=13, y=2: closest beacon is at x=15, y=3
Sensor at x=12, y=14: closest beacon is at x=10, y=16
Sensor at x=10, y=20: closest beacon is at x=10, y=16
Sensor at x=14, y=17: closest beacon is at x=10, y=16
Sensor at x=8, y=7: closest beacon is at x=2, y=10
Sensor at x=2, y=0: closest beacon is at x=2, y=10
Sensor at x=0, y=11: closest beacon is at x=2, y=10
Sensor at x=20, y=14: closest beacon is at x=25, y=17
Sensor at x=17, y=20: closest beacon is at x=21, y=22
Sensor at x=16, y=7: closest beacon is at x=15, y=3
Sensor at x=14, y=3: closest beacon is at x=15, y=3
Sensor at x=20, y=1: closest beacon is at x=15, y=3"
            .to_owned(),
        10,
        20,
    );
    let (input, line, xy_max): (_, i64, i64) = read_to_string("input.txt")
        .map(|input| (input, 2_000_000, 4_000_000))
        .unwrap_or(input);
    let sensors = input.lines().map(Sensor::from).collect::<Vec<_>>();
    let mut covered: HashSet<_> = sensors
        .iter()
        .flat_map(|s| {
            let diff = line.abs_diff(s.y);
            if diff > s.distance {
                return RangeInclusive::new(1, 0);
            }
            let d = s.distance - diff;
            (s.x - d as i64)..=(s.x + d as i64)
        })
        .collect();
    for sensor in &sensors {
        if sensor.y_beacon == line {
            covered.remove(&sensor.x_beacon);
        }
    }
    println!("{} positions can't contain a beacon", covered.len());
    'outer: for y in 0..=xy_max {
        let mut ranges = vec![];
        for sensor in &sensors {
            let diff = y.abs_diff(sensor.y);
            if diff > sensor.distance {
                continue;
            }
            let d = sensor.distance - diff;
            ranges.push((sensor.x - d as i64)..=(sensor.x + d as i64));
        }
        ranges.sort_unstable_by(|r1, r2| r1.start().cmp(r2.start()));
        let mut max_x = 0;
        for range in ranges {
            if max_x + 1 < *range.start() {
                let x = max_x + 1;
                println!(
                    "found beacon at {x},{y} with frequency {}",
                    x * 4_000_000 + y
                );
                break 'outer;
            }
            max_x = max_x.max(*range.end());
        }
    }
}
